/*

*************************************************************************

ArmageTron -- Just another Tron Lightcycle Game in 3D.
Copyright (C) 2000  Manuel Moos (manuel@moosnet.de)

**************************************************************************

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
***************************************************************************

*/

#ifndef ArmageTron_CONFIGURATION_H
#define ArmageTron_CONFIGURATION_H

#include "tList.h"
#include "tString.h"
#include "tLinkedList.h"
#include "tConsole.h"
#include <iostream.h>
#include <ctype.h>


class tConfItemBase:public tListItem<tConfItemBase>{
  friend class tCheckedPTRBase;

  int id;
protected:
  const tString title;
  const tString help;
  bool changed;

  static tConfItemBase* s_ConfItemAnchor;
  //static tConfItemBase* Anchor(){return dynamic_cast<tConfItemBase *>(s_ConfItemAnchor);}
public:
  static bool printChange;

  tConfItemBase(const char *title,const char *help);
  virtual ~tConfItemBase();

  static void SaveAll(ostream &s);
  static void LoadAll(istream &s);

  virtual void ReadVal(istream &s)=0;
  virtual void WriteVal(ostream &s)=0;

  virtual void WasChanged(){} // what to do if a read changed the thing

  virtual bool Writable(){return true;}

  virtual bool Save(){return true;}
};

// Arg! Msvc++ can't handle bool IO...
#ifdef _MSC_VER
inline istream & operator >> (istream &s,bool &b){
 int x;
 s >> x;
 b=(x!=0);
 return s;
}

inline ostream & operator << (ostream &s,bool b){
  if (b)
	return s << 1;
  else
    return s << 0;
}
#endif

template<class T> class tConfItem:virtual public tConfItemBase{
protected:
  T    *target;

  tConfItem(T &t):tConfItemBase("",""),target(&t){};
public:
  tConfItem(const char *title,const char *help,T& t)
    :tConfItemBase(title,help),target(&t){}
  virtual ~tConfItem(){}
  
  virtual void ReadVal(istream &s){
    // eat whitepsace
    char c=' ';
    while(isspace(c) && c!='\n' && s.good() && !s.eof()) c=s.get();
    s.putback(c);

    T dummy=0;
    if (c!='\n' && s && !s.eof() && s.good()){
      s >> dummy;
      if (!s)
	con << "input error reading " << title << ": invalid format.\n";
      else
	if (dummy!=*target){
	  if (!Writable())
	    con << "No way. Only the server can change " << title << ".\n";
	  else{
	    if (printChange)
	      con << title << " changed from " << *target << " to " << dummy << ".\n";
	    *target=dummy;
	    changed=true;
	  }
	}
    }
    else
      con << title << " is currently set to " << *target << ".\n";
    
    // read the rest of the line
     c=' ';
    while(c!='\n' && s.good() && !s.eof()) c=s.get();
  }
  
  virtual void WriteVal(ostream &s){
    s << *target;
  }
};

template<class T> class tSettingItem:public tConfItem<T>{
public:
  tSettingItem(const char *title,const char *help,T& t)
    :tConfItemBase(title,help),tConfItem<T>(t){}
  virtual ~tSettingItem(){}

  virtual bool Save(){return false;}
};


class tConfItemLine:public tConfItem<tString>, virtual public tConfItemBase{
public:
  tConfItemLine(const char *title,const char *help,tString &s)
    :tConfItem<tString>(title,help,s),tConfItemBase(title,help){};
  virtual ~tConfItemLine(){}

  virtual void ReadVal(istream &s);
  virtual void WriteVal(ostream &s);
};

typedef void CONF_FUNC(istream &s);

class tConfItemFunc:public tConfItemBase{
  CONF_FUNC *f;
public:
  tConfItemFunc(const char *title,const char *help,CONF_FUNC *func);
  virtual ~tConfItemFunc();

  virtual void ReadVal(istream &s);
  virtual void WriteVal(ostream &s);

  virtual bool Save();
};

void st_LoadConfig();
void st_SaveConfig();

extern bool st_FirstUse;

#endif

